from wordlette.cms.forms import Form
from wordlette.cms.states.setup.enums import SetupCategory, SetupStatus
from wordlette.cms.states.setup.route_types import SetupRoute
from wordlette.cms.themes import Template
from wordlette.configs.managers import ConfigManager
from wordlette.core.app import AppSetting
from wordlette.forms.exceptions import FormValidationError
from wordlette.forms.field_types import Link, SubmitButton, CheckBoxField, TextField
from wordlette.requests import Request
from wordlette.utils.dependency_injection import inject


class CreateSettingsFileForm(Form):
    site_name: str @ TextField(
        placeholder='ex. "Bob\'s Blog" or "Paul\'s Portfolio"',
        label="What is your website's name?",
    )
    domain_name: str @ TextField(
        placeholder='ex. "bobsblog.com" or "www.paulsportfolio.com"',
        label="What is your website's domain name?",
    )
    force_https: bool @ CheckBoxField(
        classes=["label-inline"],
        checked=True,
        label="Force HTTPS",
    )

    buttons = (
        Link("Back", href="/"),
        SubmitButton("Next", type="submit"),
    )

    def validate_type_str_isnt_empty(self, value: str):
        if not value.strip():
            raise ValueError("Field cannot be empty")


class ConfigureDatabase(SetupRoute, setup_category=SetupCategory.Database):
    path = "/configure-database"

    async def setup_status(
        self,
        config: ConfigManager @ inject = None,
        settings_filename: str @ AppSetting("settings-filename") = None,
        working_directory: str @ AppSetting("working-directory") = None,
    ) -> SetupStatus:
        # if config.find_config_file(settings_filename, working_directory):
        #     return SetupStatus.Complete

        return SetupStatus.Ready

    async def get_setup_page(self, _: Request.Get):
        return self._create_page_template(
            form=CreateSettingsFileForm,
        )

    async def create_settings_file(
        self,
        form: CreateSettingsFileForm,
        config: ConfigManager @ inject,
        settings_filename: str @ AppSetting("settings-filename"),
        working_directory: str @ AppSetting("working-directory"),
    ):
        path = config.write_config_file(
            settings_filename,
            working_directory,
            {
                "site": {
                    "domain": form.domain_name,
                    "https": form.force_https,
                    "name": form.site_name,
                }
            },
        )
        subtitle = "Your Settings Have Been Saved"
        return Template(
            "create-settings-file.html",
            title=f"Wordlette: {subtitle}",
            heading="Setup",
            subtitle=subtitle,
            created_settings_file=True,
            file_path=path,
            next_page_url=self.get_next_step_url(),
        )

    async def handle_form_validation_errors(
        self, validation_errors: FormValidationError
    ):
        return self._create_page_template(
            form=validation_errors.form,
        )

    def _create_page_template(self, **kwargs):
        subtitle = "Configure Database"
        return Template(
            "create-settings-file.html",
            title=f"Wordlette: {subtitle}",
            heading="Setup",
            subtitle=subtitle,
            **kwargs,
        )



from pathlib import Path

import jinja2

import wordlette.cms.themes.theme_managers as tm
from wordlette.utils.options import Value, Null


class JinjaTemplateLoader(jinja2.BaseLoader):
    def __init__(self, theme_manager: "tm.ThemeManager"):
        self.theme_manager = theme_manager

    def get_source(self, environment, template):
        match self.theme_manager.find_template(template):
            case Value(template_location):
                return self._load_template(template_location)

            case Null(exception):
                raise exception

    def _load_template(self, template_location: Path):
        tm.logger.debug(
            f"Found template {template_location.name!r} in {template_location.parent.resolve()}"
        )

        last_mtime = template_location.stat().st_mtime

        def is_current():
            return template_location.stat().st_mtime == last_mtime

        with template_location.open() as template_file:
            return template_file.read(), str(template_location), is_current



import logging
from pathlib import Path
from typing import Annotated, Any

import jinja2
from bevy import dependency

from wordlette.cms.exceptions import TemplateNotFound, ThemeNotFound
from wordlette.cms.themes.template_loaders import JinjaTemplateLoader
from wordlette.utils.options import Option, Value, Null

logger = logging.getLogger("Theming")


class ThemeManager:
    wordlette_res: Annotated[Path, "package-resources"] = dependency()

    def __init__(self):
        self._theme = None
        self._secondary_themes = []
        self._jinja = jinja2.Environment(
            loader=JinjaTemplateLoader(self), autoescape=True
        )

    @property
    def default_theme(self) -> Path:
        return self.wordlette_res / "themes" / "default"

    def find_template(self, template_name: str) -> Option[Path]:
        for theme in self._iterate_themes():
            template_location = theme / template_name
            if template_location.exists():
                return Value(template_location)

        return Null(
            TemplateNotFound(
                f"Could not find a template with the name {template_name!r} in any of the active themes."
            )
        )

    def set_theme(self, theme_location: Path):
        if not theme_location.exists():
            raise ThemeNotFound(
                f"Could not find a theme at {theme_location.resolve()}."
            )

        self._theme = theme_location
        logger.debug(f"Set theme to {theme_location.resolve()}")

    def add_secondary_theme(self, theme_location: Path):
        if not theme_location.exists():
            raise ThemeNotFound(
                f"Could not find a theme at {theme_location.resolve()} to use as a secondary theme."
            )

        self._secondary_themes.append(theme_location)

    def render_template(self, template_name: str, **context: Any) -> str:
        return self._jinja.get_template(template_name).render(**context)

    def _iterate_themes(self):
        if self._theme:
            yield self._theme

        yield from self._secondary_themes
        yield self.default_theme



from enum import Enum


class SetupStatus(str, Enum):
    Ready = "ready"
    Waiting = "waiting"
    Complete = "complete"


class SetupCategory(str, Enum):
    NoCategory = "no_category"
    Config = "setup_config"
    Database = "setup_database"
    Admin = "setup_admin"
    General = "setup_general"



import inspect
from functools import wraps
from types import MethodType, UnionType
from typing import (
    Callable,
    Generic,
    ParamSpec,
    TypeVar,
    get_args,
    Type,
)

from bevy import get_repository
from starlette.datastructures import FormData
from starlette.responses import Response
from starlette.types import Scope, Receive, Send

from wordlette.core.exceptions import (
    MissingRoutePath,
    NoRouteHandlersFound,
    CannotHandleInconsistentTypes,
)
from wordlette.forms import Form
from wordlette.requests import Request
from wordlette.routers import Router
from wordlette.routes.exception_contexts import ExceptionHandlerContext
from wordlette.routes.exceptions import NoCompatibleFormError
from wordlette.routes.method_collections import MethodsCollection
from wordlette.routes.route_metadata import RouteMetadataSetup
from wordlette.utils.apply import apply
from wordlette.utils.dependency_injection import AutoInject
from wordlette.utils.match_types import TypeMatchable
from wordlette.utils.options import Option

P = ParamSpec("P")
RequestType = TypeVar("RequestType", bound=Request)
ExceptionType = TypeVar("ExceptionType", bound=Exception)


def get_functions(cls):
    for name, attr in vars(cls).items():
        if (
            callable(attr)
            and not isinstance(attr, MethodType)
            and not name.startswith("_")
        ):
            yield attr


def get_handler_type(function):
    arg_spec = inspect.getfullargspec(unwrap(function))
    arg_type = (
        arg_spec.annotations.get(arg_spec.args[1]) if len(arg_spec.args) > 1 else None
    )
    return Option.Value(arg_type) if arg_type else Option.Null()


def never_abstract(func: Callable[P, None]) -> Callable[P, None]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        if args[0].__metadata__.abstract:
            return

        return func(*args, **kwargs)

    return wrapper


def unwrap(function):
    while hasattr(function, "__wrapped__"):
        function = function.__wrapped__

    return function


class RouteMCS(type):
    methods: MethodsCollection
    name: str
    path: str

    def __repr__(cls):
        if hasattr(cls, "path"):
            path, name, methods = cls.path, cls.name, cls.methods.names
            return f"<Route:{cls.__name__} {path=} {name=} {methods=}>"

        return f"<Route:Abstract:{cls.__name__}>"


class Route(
    Generic[RequestType],
    RouteMetadataSetup,
    TypeMatchable,
    AutoInject,
    metaclass=RouteMCS,
):
    """The route type handles all the magic that allows routes to be defined without any decorator boilerplate. It
    provides the instrumentation necessary for simple request handling by method type using type annotations. It also
    provides a simple exception handling mechanism using the same approach.

    To declare a request handler create an async method with whatever name you think is appropriate. Then all you have
    to do is set the type annotion of the first parameter to the request type your handler should handle. For example:
    ```python
    class MyRoute(Route):
        async def index(self, request: Request.Get):
            return PlainTextResponse("Hello World!")
    ```
    Handling errors is just as simple and follows the exact same process. The exception type will match all exceptions
    that pass a subclass check, so it should function the same as an equivalent try/except. For example:
    ```python
    class MyRoute(Route):
        ...
        async def handle_exception(self, error: Exception):
            return PlainTextResponse("An error occurred!")
    ```
    """

    methods: MethodsCollection
    name: str
    path: str

    def __init_subclass__(cls, **kwargs):
        """Set up the route class building the route meta, scanning for request & error handlers, and ensuring that
        concrete routes have a path and route handlers."""
        super().__init_subclass__(**kwargs)
        cls._setup_route()
        cls._scan_for_handlers()
        cls._setup_form_handlers()
        cls._setup_methods()
        cls._validate_route_object()
        cls._register_route()

    async def __call__(self, scope, receive, send):
        """Handle a request, calling the appropriate handler function and capturing any handleable exceptions."""
        async with ExceptionHandlerContext(self) as error_handler:
            await self._handle(scope, receive, send)

        if error_handler:
            await error_handler(scope, receive, send)

    async def process_forms(self, request: Request) -> Response:
        """Process all forms in the request body and return the response."""
        form_data = await request.form()
        match self._find_best_form_handler(form_data, type(request)):
            case Option.Value((form_type, handler)):
                return await handler(self, form_type.create_from_form_data(form_data))

            case _:
                raise NoCompatibleFormError(
                    f"Form handler for {request.name} request not found for {self.__class__.__qualname__}"
                )

    def _find_best_form_handler(
        self, form_data: FormData, method_type: Type[Request]
    ) -> Option[tuple[Type[Form], Callable[[Form], Response]]]:
        handler, highest_field_count = Option.Null(), 0
        for form, handler in self.__metadata__.form_handlers.items():
            if not form.can_handle_method(method_type):
                continue

            supported_fields = form.count_matching_fields(form_data)
            if supported_fields > highest_field_count:
                handler, highest_field_count = (
                    Option.Value((form, handler)),
                    supported_fields,
                )

            if supported_fields == len(form_data):
                break

        return handler

    async def _handle(self, scope: Scope, receive: Receive, send: Send):
        request = Request.factory(scope, receive, send)
        if type(request) not in self.__metadata__.request_handlers:
            raise Exception(
                f"Request type {type(request)} not handled by {self.__class__.__qualname__}"
            )

        handler_registry = self.__metadata__.request_handlers
        request_type = type(request)
        response = await handler_registry[request_type](self, request)
        await response(scope, receive, send)

    @classmethod
    def url(cls, **path_params):
        router: Router = get_repository().get(Router)
        return router.url_for(cls, **path_params)

    @classmethod
    def register_routes(cls, router: Router):
        """Register all routes in the route registry with the given router."""
        apply(router.add_route).to(cls.__metadata__.registry)

    @classmethod
    def _add_handlers(cls, function, handler_types, container):
        for handler_type in handler_types:
            container[handler_type] = function

    @classmethod
    def _register_handlers(cls, function, *handler_types):
        if all(issubclass(handler, Request) for handler in handler_types):
            handler_registry = cls.__metadata__.request_handlers

        elif all(issubclass(handler, Form) for handler in handler_types):
            handler_registry = cls.__metadata__.form_handlers

        elif all(issubclass(handler, Exception) for handler in handler_types):
            handler_registry = cls.__metadata__.error_handlers

        else:
            raise cls._create_inconsistent_handler_types_exception(handler_types)

        cls._add_handlers(function, handler_types, handler_registry)

    @staticmethod
    def _create_inconsistent_handler_types_exception(handler_types):
        types = ", ".join(
            handler_type.__qualname__ for handler_type in handler_types[:~0]
        )
        types += f", and {handler_types[~0].__qualname__}"
        raise CannotHandleInconsistentTypes(
            f"A route handler cannot be registered for inconsistent types: {types} do not share a compatible common"
            f" base type."
        )

    @classmethod
    @never_abstract
    def _register_route(cls):
        cls.__metadata__.registry.add(cls)

    @classmethod
    def _scan_for_handlers(cls):
        for function in get_functions(cls):
            handler_type = get_handler_type(function)
            match handler_type:
                case Option.Value(type() as ht) if issubclass(
                    ht, (Request, Exception, Form)
                ):
                    cls._register_handlers(function, ht)

                case Option.Value(UnionType() as ht):
                    cls._register_handlers(function, *get_args(ht))

    @classmethod
    def _setup_methods(cls):
        cls.methods = MethodsCollection(*cls.__metadata__.request_handlers)

    @classmethod
    @never_abstract
    def _setup_route(cls):
        if not hasattr(cls, "name"):
            cls.name = cls.__qualname__.casefold()

    @classmethod
    @never_abstract
    def _setup_form_handlers(cls):
        methods = {
            form_type.__request_method__ for form_type in cls.__metadata__.form_handlers
        }
        for method in methods:
            if method not in cls.__metadata__.request_handlers:
                cls._register_handlers(cls.process_forms, method)

    @classmethod
    @never_abstract
    def _validate_route_object(cls):
        if not hasattr(cls, "path"):
            raise MissingRoutePath(
                f"Route subclass {cls.__qualname__} is missing a path attribute."
            )

        if not cls.methods:
            raise NoRouteHandlersFound(
                f"Route subclass {cls.__qualname__} has no request handlers."
            )



from collections import defaultdict
from inspect import get_annotations
from itertools import chain
from types import UnionType, MethodType
from typing import (
    Annotated,
    Any,
    Callable,
    Hashable,
    Sequence,
    Type,
    TypeAlias,
    TypeVar,
    get_args,
    get_origin,
    Iterable,
    ParamSpec,
)

from starlette.datastructures import FormData

import wordlette.forms
from wordlette.forms.exceptions import FormValidationError
from wordlette.forms.field_types import SubmitButton, Button
from wordlette.forms.fields import Field, NotSet, not_set
from wordlette.forms.views import FormView
from wordlette.requests import Request
from wordlette.utils.contextual_methods import contextual_method

P = ParamSpec("P")
F = TypeVar("F", bound="Form")
T = TypeVar("T")
Validator: TypeAlias = Callable[[T], None]


class FieldScanner:
    def __init__(self, form: "Type[Form]"):
        self.form = form
        self.fields = {}

    def scan(self):
        for name, type_hint in get_annotations(self.form).items():
            if not name.startswith("_"):
                self.fields[name] = self._setup_form_field(name, type_hint)

    def _setup_form_field(self, name: str, hint: Type[Field] | T) -> Field:
        params = {"type_hint": hint}
        if hasattr(self.form, name):
            params["value"] = params["default"] = getattr(self.form, name)

        construct: Callable[[P], Field] = Field.from_type
        if _is_annotated_field(hint):
            params["type_hint"], field = get_args(hint)
            construct = field.set_missing

            if isinstance(field, type):
                return field(**params)

            html_name = name.replace("_", "-").casefold()
            if field.label is not not_set:
                params["id"] = field.attrs.get("name") or html_name

            params["name"] = field.attrs.get("id") or html_name

        if get_origin(params["type_hint"]) is UnionType:
            args = get_args(params["type_hint"])
            if len(args) > 1 and not isinstance(None, args[1]):
                raise TypeError(
                    f"Form field's may not have more than one annotated type. Got {len(args)} annotations."
                )

            params["type_hint"] = args[0]
            params["optional"] = type(None) in args
            params.setdefault("default", None)

        return construct(**params)


def _is_annotated_field(hint) -> bool:
    if get_origin(hint) is not Annotated:
        return False

    _, field = get_args(hint)
    if isinstance(field, type):
        return issubclass(field, Field)

    return isinstance(field, Field)


class ValidatorScanner:
    def __init__(self, cls: "Type[Form]", field_names: Sequence[str]):
        self.cls = cls
        self.field_names = field_names
        self.validators = defaultdict(list)
        self.type_validators = defaultdict(list)

    def scan(self):
        for name in vars(self.cls):
            attr = getattr(self.cls, name)
            if not callable(attr):
                continue

            if name.startswith("validate_type"):
                self._process_type_validator(attr)

            elif name.startswith("validate_"):
                self._process_validator(attr, name)

    def _get_matching_field_name(self, name: str) -> str | None:
        for field_name in self.field_names:
            if name.startswith(f"validate_{field_name}"):
                return field_name

    def _process_validator(self, validator: Validator, name: str):
        field_name = self._get_matching_field_name(name)
        self.validators[field_name].append(validator)

    def _process_type_validator(self, validator: Validator):
        for type_ in self._get_type_validator_types(validator):
            self.type_validators[type_].append(validator)

    def _get_type_validator_types(self, validator: Validator) -> list[Type]:
        annotations = get_annotations(validator)
        annotation = next(iter(annotations.values()))
        origin = get_origin(annotation)
        if origin is UnionType:
            args = get_args(annotation)
            if args[~0] is None:
                args = args[:~0]

            types = list(args)

        else:
            types = [annotation]

        return [
            get_args(type_)[0] if get_origin(type_) is Annotated else type_
            for type_ in types
        ]


class Form:
    __form_fields__: dict[str, Field]
    __form_field_names__: dict[str, str]
    __validators__: dict[str, list[Validator]] = {}
    __type_validators__: dict[Type, list[Validator]] = {}
    __request_method__: Type[Request]
    __form_view_type__: Type[FormView]

    buttons: Iterable[Button] = (SubmitButton("Submit"),)

    def __init_subclass__(
        cls,
        method: Type[Request] = Request.Post,
        field_scanner: Type[FieldScanner] | None = None,
        validator_scanner: Type[ValidatorScanner] | None = None,
        view: Type[FormView] | None = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)
        cls.__request_method__ = method
        cls.__form_view_type__ = view or getattr(cls, "__form_view_type__", FormView)
        cls._setup_form_fields(field_scanner or FieldScanner)
        cls._setup_buttons()
        cls._setup_validators(validator_scanner or ValidatorScanner)

    def __init__(self, *args, **kwargs):
        self.__field_values__ = {}
        self._load_fields(*args, **kwargs)

        if errors := self._validate_fields():
            raise self._create_validation_exception(errors)

    def get_field_value(self, name: str) -> Any | NotSet:
        field_name = self.__form_field_names__[name]
        field = self.__form_fields__[field_name]
        return self.__field_values__.get(field_name, field.default)

    @contextual_method
    def view(self) -> FormView:
        return self.__form_view_type__(
            self.__form_fields__,
            self.buttons,
            self.__field_values__,
            self._validate_fields(),
        )

    @view.classmethod
    def view(cls) -> FormView:
        return cls.__form_view_type__(
            cls.__form_fields__,
            cls.buttons,
            {},
            {},
        )

    def _load_fields(self, *args, **kwargs):
        required_fields = [
            field for field in self.__form_fields__.values() if field.required
        ]
        positional_args = list(reversed(args))
        for name, field in self.__form_fields__.items():
            if name in kwargs:
                value = kwargs[name]
            elif positional_args:
                value = positional_args.pop()
            elif field in required_fields:
                raise TypeError(
                    f"{type(self).__qualname__} is missing a value for the required field {name!r}"
                )
            else:
                continue

            self.__field_values__[name] = field.convert(value)

    def _validate_fields(self) -> dict[str, Exception]:
        validation_errors = {}
        for name, value in self.__field_values__.items():
            try:
                self._validate_field(name, value)
            except Exception as e:
                validation_errors[self.__form_fields__[name].name] = e

        return validation_errors

    def _validate_field(self, name: str, value: Any):
        field = self.__form_fields__[name]
        for validator in chain([field.validate], self.__validators__.get(name, ())):
            if not isinstance(validator, MethodType):
                validator = MethodType(validator, self)

            validator(value)

        for type_, validators in self.__type_validators__.items():
            if isinstance(value, type_):
                for validator in validators:
                    if not isinstance(validator, MethodType):
                        validator = MethodType(validator, self)

                    validator(value)

    @classmethod
    def _setup_buttons(cls):
        if isinstance(cls.buttons, Button):
            cls.buttons = (cls.buttons,)

    @classmethod
    def _setup_form_fields(cls, scanner_type: Type[FieldScanner]):
        scanner = scanner_type(cls)
        scanner.scan()

        cls.__form_fields__ = scanner.fields
        cls.__form_field_names__ = {}
        for name, field in scanner.fields.items():
            setattr(cls, name, field)
            cls.__form_field_names__[field.name] = name

    @classmethod
    def _setup_validators(cls, scanner_type: Type[ValidatorScanner]):
        scanner = scanner_type(cls, cls.__form_fields__)
        scanner.scan()

        _merge_dicts_of_lists(cls.__validators__, scanner.validators)
        _merge_dicts_of_lists(cls.__type_validators__, scanner.type_validators)

    @classmethod
    def add_type_validator(
        cls, type_: Type[T], validator: Validator | None = None
    ) -> Callable[[Validator], Validator] | None:
        if not validator:

            def decorator(v: Validator) -> Validator:
                cls.add_type_validator(type_, v)
                return v

            return decorator

        cls.__type_validators__.setdefault(type_, []).append(lambda _, v: validator(v))

    @classmethod
    def can_handle_method(cls, method: Type[Request]) -> bool:
        return issubclass(method, cls.__request_method__)

    @classmethod
    def create_from_form_data(cls: Type[F], data: FormData) -> F:
        return cls(
            **{
                cls.__form_field_names__[name]: value
                for name, value in data.items()
                if name in cls.__form_field_names__
            }
        )

    @classmethod
    def count_matching_fields(cls, data: FormData) -> int:
        if len(data) < len(cls.__form_fields__) or any(
            name not in data for name in cls.__form_field_names__
        ):
            return 0

        return len(cls.__form_fields__)

    def _create_validation_exception(self, errors):
        error_count = "an error"
        if len(errors) > 1:
            error_count = f"{len(errors):,} errors"

        return FormValidationError(
            f"Form validation failed with {error_count}.", errors, self
        )


def _merge_dicts_of_lists(
    base_dict: dict[Hashable, list[Any]], update_dict: dict[Hashable, list[Any]]
):
    for key, value in update_dict.items():
        base_dict.setdefault(key, []).extend(value)


@Form.add_type_validator(int)
def validate_type_int_fits_in_a_bigint(value: int):
    if value > 2**63 - 1:
        raise wordlette.forms.ValidationError(
            "Integers must not be greater than 2^63 - 1 (largest signed 64-bit integer/bigint)"
        )

    if value < -(2**63):
        raise wordlette.forms.ValidationError(
            "Integers must not be less than -2^63 (smallest signed 64-bit integer/bigint)"
        )



from inspect import get_annotations
from typing import (
    Type,
    TypeVar,
    cast,
    Annotated,
    Any,
    Callable,
    ParamSpec,
    Iterator,
    Self,
)

from wordlette.forms.elements import Element
from wordlette.utils.annotated_aggregator import AnnotatedAggregator
from wordlette.utils.sentinel import sentinel

NotSet, not_set = sentinel("NotSet")
P = ParamSpec("P")
T = TypeVar("T")


class ConverterScanner:
    def scan(self, cls: Type) -> dict[Type, str]:
        return dict(self._scan(cls))

    def _scan(self, cls: Type) -> Iterator[tuple[Type, str]]:
        for name, return_type in self._get_converter_return_types(cls):
            if return_type is not None:
                yield return_type, name

    def _get_converter_return_types(self, cls) -> Iterator[tuple[str, Type[T] | None]]:
        for name, method in vars(cls).items():
            if callable(method) and self._valid_name(name):
                yield name, self._get_return_type(method)

    def _get_return_type(self, method: Callable[[P], Type[T]]) -> Type[T] | None:
        return get_annotations(method).pop("return", None)

    def _valid_name(self, name: str) -> bool:
        return name.startswith("convert_")


class FieldMCS(type):
    def __rmatmul__(cls, other: T) -> T:
        return cast(T, Annotated[other, cls])


class Field(metaclass=FieldMCS):
    __converters__: dict[Type, str] = {}

    def __init_subclass__(
        cls, converter_scanner: Type[ConverterScanner] = ConverterScanner, **kwargs
    ):
        super().__init_subclass__(**kwargs)
        cls.__converters__ = cls._scan_for_converters(converter_scanner)

    @classmethod
    def _scan_for_converters(
        cls, scanner_type: Type[ConverterScanner]
    ) -> dict[Type, str]:
        return getattr(cls, "__converters__", {}) | scanner_type().scan(cls)

    def __init__(
        self,
        type_hint: Type[T] | None = None,
        default: T | NotSet = not_set,
        label: str | Element | NotSet = not_set,
        **attrs,
    ):
        self.attrs = attrs
        self.default = default
        self.label = label
        self.optional = False
        self.type_hint = type_hint

    def __rmatmul__(self, other: T) -> T:
        return cast(T, AnnotatedAggregator[other, self])

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return instance.get_field_value(self.name)

    @property
    def name(self) -> str:
        return self.attrs.get("name", "")

    @property
    def required(self) -> bool:
        return not self.optional and self.default is not_set

    def compose(self, value: Any | NotSet = not_set) -> Element:
        from wordlette.forms.elements import InputElement

        params = self.attrs.copy()
        if value is not not_set:
            params["value"] = value

        if self.required:
            params["required"] = True

        return InputElement(**params)

    def compose_label(self) -> Element | None:
        from wordlette.forms.elements import LabelElement

        match self.label:
            case LabelElement():
                return self.label

            case str() as body:
                return LabelElement(
                    body, **self._filter_and_clean_params(for_=self.attrs["id"])
                )

            case _:
                return None

    def convert(self, value: Any) -> T:
        converter_name = self.__converters__.get(self.type_hint)
        converter = getattr(self, converter_name) if converter_name else self.type_hint
        return converter(value)

    def set_missing(self, **kwargs) -> Self:
        self.default = kwargs.pop("default", self.default)
        self.optional = kwargs.pop("optional", self.optional)
        self.type_hint = kwargs.pop("type_hint", self.type_hint)

        self.attrs |= {k: v for k, v in kwargs.items() if k not in self.attrs}
        return self

    def validate(self, value: T):
        return

    def _filter_and_clean_params(self, **kwargs):
        return {k.rstrip("_"): v for k, v in kwargs.items() if v is not not_set}

    @classmethod
    def from_type(cls, type_hint: "Type[Field]", **config) -> "Field":
        return cls(type_hint=type_hint, **config)



from typing import Type, cast


class Sentinel:
    def __new__(cls):
        inst = super().__new__(cls)
        cls.__new__ = lambda _: inst
        return inst

    def __repr__(self):
        return f"<{type(self).__name__}>"


def sentinel(name: str) -> tuple[Type[Sentinel], Sentinel]:
    """Create a new sentinel type."""

    cls = cast(Type[Sentinel], type(name, (Sentinel,), {}))
    return cls, cls()



import re
from datetime import datetime, timedelta, date, time
from decimal import Decimal
from enum import Enum
from typing import Any, Annotated

from wordlette.forms import Field
from wordlette.forms.elements import ButtonElement, AElement
from wordlette.forms.fields import NotSet, not_set


class Week:
    def __init__(self, year: int, week: int):
        if week < 1 or week > 52:
            raise ValueError("Week must be between 1 and 52")

        self.year = year
        self.week = week


class CaptureType(str, Enum):
    USER = "user"
    ENVIRONMENT = "environment"


class FormEncodingType(str, Enum):
    APPLICATION_X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded"
    MULTIPART_FORM_DATA = "multipart/form-data"
    TEXT_PLAIN = "text/plain"


class Label:
    def __init__(self, text: str, for_: str = ""):
        self.text = text
        self.for_ = for_


class InputField(Field):
    type: str

    def __init__(
        self,
        *,
        disabled: bool | NotSet = not_set,
        value: Any | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(
            **self._filter_and_clean_params(
                type_=self.type,
                disabled=disabled,
                value=value,
                **kwargs,
            )
        )


class TextField(InputField):
    type = "text"

    def __init__(
        self,
        *,
        max_length: int | NotSet = not_set,
        min_length: int | NotSet = not_set,
        pattern: str | NotSet = not_set,
        placeholder: str | NotSet = not_set,
        readonly: bool | NotSet = not_set,
        size: int | NotSet = not_set,
        spellcheck: bool | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(
            maxlength=max_length,
            minlength=min_length,
            pattern=pattern,
            placeholder=placeholder,
            readonly=readonly,
            size=size,
            spellcheck=spellcheck,
            **kwargs,
        )


class ButtonField(InputField):
    type = "button"


class CheckBoxField(InputField):
    type = "checkbox"

    def __init__(self, checked: bool | NotSet = not_set, **kwargs):
        super().__init__(checked=checked, **kwargs)

    @property
    def required(self):
        return False

    def convert_to_bool(self, value: str) -> bool:
        if "value" in self.attrs:
            return value == self.attrs["value"]

        return value.casefold() == "on"


class ColorField(InputField):
    type = "color"

    def __init__(self, value: str | NotSet = not_set, **kwargs):
        if value is not not_set and not re.match(r"^#[0-9a-fA-F]{6}$", value):
            raise ValueError("Color value must be a valid hex color code")

        super().__init__(value=value, **kwargs)


class DateField(InputField):
    type = "date"

    def __init__(
        self,
        *,
        min_date: date | NotSet = not_set,
        max_date: date | NotSet = not_set,
        step: Annotated[int | NotSet, "A step set in days"] = not_set,
        **kwargs,
    ):
        super().__init__(min=min_date, max=max_date, step=step, **kwargs)


class DateTimeField(InputField):
    type = "datetime-local"

    def __init__(
        self,
        *,
        min_datetime: datetime | NotSet = not_set,
        max_datetime: datetime | NotSet = not_set,
        step: timedelta | NotSet = not_set,
        value: datetime | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(
            max=max_datetime,
            min=min_datetime,
            step=step,
            value=value,
            **kwargs,
        )


class EmailField(TextField):
    type = "email"

    def __init__(self, *, pattern: str | NotSet = not_set, **kwargs):
        super().__init__(pattern=pattern, **kwargs)


class MultipleEmailField(EmailField):
    def __init__(self, *, pattern: str | NotSet = not_set, **kwargs):
        super().__init__(pattern=pattern, **kwargs)


class FileField(InputField):
    type = "file"

    def __init__(
        self,
        *,
        accept: str | NotSet = not_set,
        capture: CaptureType | str | NotSet = not_set,
        **kwargs,
    ):
        if "value" in kwargs:
            raise ValueError("You cannot set the value of a file field")

        super().__init__(
            accept=accept,
            capture=capture,
            **kwargs,
        )


class MultipleFileField(FileField):
    def __init__(
        self,
        *,
        accept: str | NotSet = not_set,
        capture: CaptureType | str | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(accept=accept, capture=capture, **kwargs)


class HiddenField(InputField):
    type = "hidden"


class MonthField(InputField):
    type = "month"

    def __init__(
        self,
        *,
        min_month: date | NotSet = not_set,
        max_month: date | NotSet = not_set,
        step: Annotated[int | NotSet, "A step set in months"] = not_set,
        **kwargs,
    ):
        super().__init__(max=max_month, min=min_month, step=step, **kwargs)


class NumberField(InputField):
    type = "number"

    def __init__(
        self,
        min_value: int | NotSet = not_set,
        max_value: int | NotSet = not_set,
        step: int | NotSet = not_set,
        value: int | NotSet = not_set,
        **kwargs,
    ):
        self._validate(min_value, max_value, step, value)
        super().__init__(max=max_value, min=min_value, step=step, value=value, **kwargs)

    def _validate(self, min_value, max_value, step, value):
        if step is not not_set and step <= 0:
            raise ValueError("step must be greater than 0")

        if (
            min_value is not not_set
            and max_value is not not_set
            and min_value >= max_value
        ):
            raise ValueError("min_value must be less than max_value")

        if min_value is not not_set and not isinstance(min_value, (int, Decimal)):
            raise TypeError("min_value must be an integer or Decimal")

        if max_value is not not_set and not isinstance(max_value, (int, Decimal)):
            raise TypeError("max_value must be an integer or Decimal")

        if value is not not_set and not isinstance(value, (int, Decimal)):
            raise TypeError("value must be an integer or Decimal")


class PasswordField(TextField):
    type = "password"


class RadioButtonField(InputField):
    type = "radio"

    def __init__(self, *, value: Any, checked: bool | NotSet = not_set, **kwargs):
        super().__init__(value=value, checked=checked, **kwargs)

    def convert_to_bool(self, value: str) -> bool:
        return value == self.value


class RangeField(InputField):
    type = "range"

    def __init__(
        self,
        *,
        min_value: int | NotSet = not_set,
        max_value: int | NotSet = not_set,
        step: int | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(max=max_value, min=min_value, step=step, **kwargs)


class SearchField(TextField):
    type = "search"


class SubmitField(InputField):
    type = "submit"

    def __init__(
        self,
        *,
        form_action: Annotated[
            str | NotSet, "Populates the HTML formaction attribute"
        ] = not_set,
        form_encoding: Annotated[
            FormEncodingType | NotSet, "Populates the HTML formenctype attribute"
        ] = not_set,
        form_method: Annotated[
            str | NotSet, "Populates the HTML formmethod attribute"
        ] = not_set,
        form_no_validate: Annotated[
            bool | NotSet, "Populates the HTML formnovalidate attribute"
        ] = not_set,
        form_target: Annotated[
            str | NotSet, "Populates the HTML formtarget attribute"
        ] = not_set,
        **kwargs,
    ):
        super().__init__(
            formaction=form_action,
            formenctype=form_encoding,
            formmethod=form_method,
            formnovalidate=form_no_validate,
            formtarget=form_target,
            **kwargs,
        )


class ImageField(SubmitField):
    type = "image"

    def __init__(
        self,
        *,
        alt: str | NotSet = not_set,
        height: int | NotSet = not_set,
        width: int | NotSet = not_set,
        src: str | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(alt=alt, height=height, width=width, src=src, **kwargs)


class TelField(TextField):
    type = "tel"


class TimeField(InputField):
    type = "time"

    def __init__(
        self,
        *,
        min_time: time | NotSet = not_set,
        max_time: time | NotSet = not_set,
        step: timedelta | NotSet = not_set,
        value: time | NotSet = not_set,
        readonly: bool | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(
            max=max_time,
            min=min_time,
            step=step,
            value=value,
            readonly=readonly,
            **kwargs,
        )


class UrlField(TextField):
    type = "url"


class WeekField(InputField):
    type = "week"

    def __init__(
        self,
        *,
        min_week: Week | NotSet = not_set,
        max_week: Week | NotSet = not_set,
        step: Annotated[int | NotSet, "A step set in weeks"] = not_set,
        readonly: bool | NotSet = not_set,
        **kwargs,
    ):
        super().__init__(
            max=max_week, min=min_week, step=step, readonly=readonly, **kwargs
        )


class Button:
    button_type: str = "button"

    def __init__(self, text: str, **attrs):
        self.text = text
        self.attrs = attrs | {"type": self.button_type}

    def compose(self):
        return ButtonElement(self.text, **self.attrs)


class SubmitButton(Button):
    button_type = "submit"


class ResetButton(Button):
    button_type = "reset"


class Link:
    def __init__(self, text: str, href: str | NotSet = not_set, **attrs):
        self.attrs = {key.rstrip("_"): value for key, value in attrs.items()}
        self.text = text

        if href is not not_set:
            self.attrs["href"] = href

    def compose(self):
        return AElement(self.text, **self.attrs)



from wordlette.base_exceptions import BaseWordletteException


class NoCompatibleFormError(BaseWordletteException):
    """Raised when no compatible form is found for the request."""

    name = "No Compatible Form Found"



import wordlette.forms
from wordlette.base_exceptions import BaseWordletteException


class FormValidationError(BaseWordletteException):
    """Raised when a form fails to validate."""

    def __init__(
        self, message: str, errors: dict[str, Exception], form: "wordlette.forms.Form"
    ):
        super().__init__(message)
        self.errors = errors
        self.form = form

        for field_name, error in errors.items():
            self.add_note(f"{field_name} - {error}")

    @property
    def name(self):
        return f"Form Validation Error{'s' if len(self.errors) > 1 else ''}"



from asyncio import iscoroutinefunction
from functools import wraps, partial
from inspect import signature, get_annotations
from typing import TypeVar, Callable, get_origin, Annotated, get_args

from bevy import get_repository
from bevy.injectors.functions import _get_function_namespace

from wordlette.at_annotateds import AtAnnotation

T = TypeVar("T", bound=Callable)


def inject_dependencies(func: T) -> T:
    sig = signature(func)
    annotations = _get_annotations(func)
    # Determine which parameters have a declared inject
    inject_parameters = {
        name: annotations[name]
        for name, parameter in sig.parameters.items()
        if name in annotations and _is_at_annotation_type(annotations[name])
    }

    @wraps(func)
    def injector(*args, **kwargs):
        params = sig.bind_partial(*args, **kwargs)
        # Get instances from the repo to fill all missing inject parameters
        params.arguments |= {
            name: get_repository().get(annotation)
            for name, annotation in inject_parameters.items()
            if name not in params.arguments
        }
        return func(*params.args, **params.kwargs)

    injector.injected_params = inject_parameters
    if iscoroutinefunction(func):

        @wraps(func)
        async def async_injector(*args, **kwargs):
            return await injector(*args, **kwargs)

        return async_injector

    return injector


class AutoInject:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for name in dir(cls):
            try:
                attr = getattr(cls, name)
            except AttributeError:
                continue

            if not name.endswith("__") and _needs_injector(attr, {cls.__name__: cls}):
                setattr(cls, name, inject_dependencies(attr))

        for name, value in get_annotations(cls).items():
            if get_origin(value) is Annotated:
                hint, annotation = get_args(value)
                if isinstance(annotation, Dependency):
                    annotation.injector = annotation.strategy(hint)
                    setattr(cls, name, annotation)


class Dependency(AtAnnotation):
    def __init__(self):
        super().__init__()
        self.injector = None

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return self.injector()

    def strategy(self, type_, *_):
        return partial(get_repository().get, type_)


def _get_annotations(func, ns=None):
    ns = (ns or {}) | _get_function_namespace(func)
    try:
        return get_annotations(func, globals=ns, eval_str=True)
    except NameError as exc:
        raise NameError(f"Couldn't process annotations on {func.__qualname__}") from exc


def _is_at_annotation_type(annotation) -> bool:
    origin = get_origin(annotation)
    if origin is not Annotated:
        return False

    return True


def _needs_injector(attr, ns=None):
    if isinstance(attr, type):
        return False

    if not callable(attr):
        return False

    if hasattr(attr, "injected_params"):
        return False

    return any(
        _is_at_annotation_type(annotation)
        for annotation in _get_annotations(attr, ns).values()
    )


inject = Dependency



import logging
from pathlib import Path
from typing import Type, TypeVar, Any, Callable, Iterable, BinaryIO

from wordlette.configs.exceptions import ConfigCannotCreateFiles
from wordlette.configs.handlers import ConfigHandler
from wordlette.core.exceptions import ConfigFileNotFound

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("Config")
T = TypeVar("T")


class ConfigManager:
    def __init__(self, handlers: Iterable[Type[ConfigHandler]] = ()):
        self._handler_extensions: dict[str, ConfigHandler] = {}
        self._handlers: list[ConfigHandler] = []
        self._config: dict[str, Any] = {}

        self._register_handlers(handlers)

    @property
    def valid_extensions(self) -> set[str]:
        return set(self._handler_extensions.keys())

    def get(
        self, key: str = "", constructor: Callable[[dict[str, Any]], T] | None = None
    ) -> T | Any:
        value = self._config[key] if key else self._config
        if constructor:
            value = constructor(**value)

        return value

    def find_config_file(self, name: str, directory: Path) -> Path | None:
        for extension in self._handler_extensions:
            path = directory / f"{name}.{extension}"
            if path.exists():
                return path

    def load_config_file(self, name: str, directory: Path):
        path = self.find_config_file(name, directory)
        if not path:
            raise ConfigFileNotFound(
                f"Could not find any config files that matched the naming scheme '{name}.*' in '{directory}'."
            )

        handler = self._handler_extensions[path.suffix[1:].casefold()]
        with path.open("rb") as open_file:
            self._config = handler.load(open_file)

        logger.debug(f"Loaded config file '{path}'.")

    def write_config_file(self, name: str, directory: Path, data: T) -> Path:
        if path := self.find_config_file(name, directory):
            logger.debug(f"Found existing config file {str(path)!r}.")
            with path.open("wb") as open_file:
                self._write_file_with_handler(
                    data,
                    open_file,
                    self._handler_extensions[path.suffix[1:].casefold()],
                )

        else:
            path = directory / f"{name}.tmp"
            with path.open("wb") as open_file:
                for handler in self._handlers:
                    logger.debug(
                        f"Attempting to create config using {type(handler).__name__}."
                    )
                    try:
                        self._write_file_with_handler(data, open_file, handler)
                    except ConfigCannotCreateFiles:
                        open_file.truncate()
                        continue
                    else:
                        extension = min(handler.extensions)
                        new_name = f"{name}.{extension}"
                        path.rename(new_name)
                        path = directory / new_name
                        break

        logger.debug(f"Wrote to config file {str(path)!r}.")

        self._config = data
        return path

    def _write_file_with_handler(
        self, data: dict[str, Any], file: BinaryIO, handler: ConfigHandler
    ):
        handler.write(data, file)

    def add_handler(self, handler_type: Type[ConfigHandler]):
        self._handlers.append(handler := handler_type())
        for extension in handler.extensions:
            self._handler_extensions[extension.casefold()] = handler

    def remove_handler(self, handler: Type[ConfigHandler]):
        for extension in handler.extensions:
            del self._handler_extensions[extension.casefold()]

    def _register_handlers(self, handlers: Iterable[Type[ConfigHandler]]):
        for handler in handlers:
            self.add_handler(handler)



from wordlette.cms.states.setup.route_types import SetupRoute
from wordlette.cms.themes import Template
from wordlette.requests import Request


class Index(SetupRoute):
    path = "/"

    async def get(self, request: Request.Get):
        next_page = await self.get_next_page()
        return Template(
            "index.html",
            title="Wordlette",
            subtitle="Setting Up Your Site",
            next_page_url=next_page.url(),
        )



from typing import Iterable

from wordlette.cms.states.setup.enums import SetupStatus, SetupCategory
from wordlette.routes import Route
from wordlette.utils.dependency_injection import inject


class CategoryController:
    def __init__(self):
        self.routes: list[SetupRoute] = []
        self.completed_routes: list[SetupRoute] = []
        self.history: list[SetupRoute] = []

    def add_route(self, route: "SetupRoute"):
        self.routes.append(route)

    def completed(self, route: "SetupRoute"):
        self.routes.remove(route)
        self.completed_routes.append(route)

    def is_done(self):
        return not self.routes

    async def get_next_route(self) -> "SetupRoute | None":
        for route in self.routes.copy():
            status = await route.setup_status()
            if status is SetupStatus.Ready:
                self.history.append(route)
                return route

            elif status is SetupStatus.Complete:
                self.completed(route)

        else:
            return self.routes[0] if self.routes else None


class SetupRouteCategoryController:
    category_order: Iterable[SetupCategory] = (
        SetupCategory.Config,
        SetupCategory.Database,
        SetupCategory.Admin,
        SetupCategory.General,
    )

    def __init__(self):
        self._category_stack = iter(self.category_order)

        self.categories = self._build_categories()
        self.current_category = self.next_category()

        self.index_route: SetupRoute | None = None
        self.completed_route: SetupRoute | None = None

    def next_category(self) -> SetupCategory:
        return next(self._category_stack, SetupCategory.NoCategory)

    def add_route(self, route: "SetupRoute"):
        if route.path == "/":
            self.index_route = route

        elif route.setup_category == SetupCategory.NoCategory:
            self.completed_route = route

        else:
            self.categories[route.setup_category].add_route(route)

    async def get_next_route(self) -> "SetupRoute":
        while self.current_category is not SetupCategory.NoCategory:
            route = await self.categories[self.current_category].get_next_route()
            if route is not None:
                return route

            self.current_category = self.next_category()

        return self.completed_route

    def _build_categories(self) -> dict[SetupCategory, CategoryController]:
        categories = {}
        for category in SetupCategory:
            categories[category] = CategoryController()

        return categories


class SetupRoute(Route):
    controller: SetupRouteCategoryController @ inject
    setup_category: SetupCategory

    class __metadata__:
        abstract = True
        registry = set()

    def __init_subclass__(
        cls, setup_category: SetupCategory = SetupCategory.General, **kwargs
    ):
        super().__init_subclass__(**kwargs)

        cls.setup_category = setup_category

    def __init__(self):
        super().__init__()
        self.controller.add_route(self)

    async def setup_status(self, *_) -> SetupStatus:
        return SetupStatus.Waiting

    async def get_next_page(self) -> "SetupRoute":
        return await self.controller.get_next_route()



from bevy import dependency, inject
from markupsafe import Markup

from wordlette.cms.themes import ThemeManager
from wordlette.forms import Form as _Form
from wordlette.forms.views import FormView as _FormView
from wordlette.utils.contextual_methods import contextual_method


class FormView(_FormView):
    def __init__(self, template: str, *args):
        super().__init__(*args)
        self.template = template

    @inject
    def render(
        self,
        __template: str | None = None,
        /,
        __tm: ThemeManager = dependency(),
        **context,
    ) -> str:
        template = __template or self.template
        return Markup(__tm.render_template(template, form=self, **context))


class Form(_Form):
    __form_template__: str = "form.html"
    __form_view_type__ = FormView

    @contextual_method
    def render(self):
        return self.view().render()

    @contextual_method
    def view(self) -> FormView:
        return self.__form_view_type__(
            self.__form_template__,
            self.__form_fields__,
            self.buttons,
            self.__field_values__,
            self._validate_fields(),
        )

    @view.classmethod
    def view(cls) -> FormView:
        return cls.__form_view_type__(
            cls.__form_template__,
            cls.__form_fields__,
            cls.buttons,
            {},
            {},
        )



from typing import Type, Any, Callable, cast

import wordlette.routes as routes
from wordlette.forms import Form
from wordlette.requests import Request


def create_metadata_type(
    *bases: "Type[RouteMetadata]" | Type, **properties: Any
) -> "Type[RouteMetadata]":
    return cast(
        Type[RouteMetadata],
        type(
            "__metadata__",
            bases,
            properties,
        ),
    )


def get_super_route(cls):
    for base in cls.__bases__:
        if base is routes.Route or issubclass(base, routes.Route):
            return base


class RouteMetadata:
    abstract: bool
    registry: "set[Type[routes.Route]]"
    error_handlers: dict[Type[Exception], Callable[[Exception], Any]]
    request_handlers: dict[Type[Request], Callable[[Request], Any]]
    form_handlers: dict[Type[Form], Callable[[Form], Any]]


class RouteMetadataSetup:
    class __metadata__(RouteMetadata):
        abstract = True
        registry = set()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if RouteMetadataSetup in cls.__bases__:
            return

        cls._setup_metadata()

    @classmethod
    def _create_new_metadata_object(cls):
        super_route = get_super_route(cls)
        cls.__metadata__ = create_metadata_type(
            super_route.__metadata__,
            abstract=False,
            request_handlers={},
            error_handlers={},
            form_handlers={},
        )

    @classmethod
    def _setup_metadata(cls):
        if cls.__metadata__ is get_super_route(cls).__metadata__:
            cls._create_new_metadata_object()

        elif not issubclass(cls.__metadata__, RouteMetadata):
            cls._transform_metadata_to_metadata_type()

    @classmethod
    def _transform_metadata_to_metadata_type(cls):
        super_route = get_super_route(cls)

        if not hasattr(cls.__metadata__, "request_handlers"):
            cls.__metadata__.request_handlers = {}

        if not hasattr(cls.__metadata__, "error_handlers"):
            cls.__metadata__.error_handlers = {}

        if not hasattr(cls.__metadata__, "form_handlers"):
            cls.__metadata__.form_handlers = {}

        cls.__metadata__ = create_metadata_type(
            cls.__metadata__, super_route.__metadata__
        )



from typing import Sequence, Any, Iterable

from markupsafe import Markup


class Element:
    tag: str
    flag_attrs = frozenset(("required", "checked", "disabled", "selected"))

    def __init__(self, *, __clone__: bool = False, **attrs):
        self.attrs = self._clean_attrs(attrs)
        self.cloned = __clone__

    def __eq__(self, other):
        if not isinstance(other, Element):
            return False

        return self.tag == other.tag and self.attrs == other.attrs

    def __repr__(self):
        return f"<{type(self).__name__} {self.tag} {self.attrs}>"

    @property
    def classes(self) -> set[str]:
        return self.attrs.get("class", set())

    def add_attr(self, **new_attrs: Any) -> "Element":
        return self.clone(**self.attrs | new_attrs)

    def add_class(self, *new_classes: str) -> "Element":
        return self.clone(**self.attrs | {"class": self.classes | set(new_classes)})

    def clone(self, **attrs) -> "Element":
        if self.cloned:
            self.attrs = attrs
            return self

        return self._create_clone(**attrs)

    def remove(self, *remove_attrs: str) -> "Element":
        return self.clone(
            **{k: v for k, v in self.attrs.items() if k not in remove_attrs}
        )

    def render(self):
        return Markup(f"<{self.tag} {self._build_attrs()} />")

    def _build_attrs(self) -> str:
        attrs_dict = self.attrs.copy()
        classes = self._process_classes(attrs_dict.pop("class", set()))
        classes |= self._process_classes(attrs_dict.pop("classes", set()))
        if classes:
            attrs_dict["class"] = " ".join(classes)

        attrs = [
            f'{k}="{Markup.escape(v)}"'
            for k, v in attrs_dict.items()
            if k not in self.flag_attrs
        ]
        flags = [
            name for name in self.flag_attrs if attrs_dict.get(name, False) is not False
        ]
        return " ".join(attrs + flags)

    def _clean_attrs(self, attrs: dict[str, Any]) -> dict[str, Any]:
        attrs = {self._clean_attr_name(k): v for k, v in attrs.items()}
        classes = self._process_classes(attrs.pop("classes", set()))
        classes |= self._process_classes(attrs.pop("class", set()))
        if classes:
            attrs["class"] = classes

        return attrs

    def _clean_attr_name(self, name: str) -> str:
        return name.rstrip("_").strip().replace("_", "-").casefold()

    def _create_clone(self, **attrs) -> "Element":
        return type(self)(__clone__=True, **attrs)

    def _process_classes(self, classes: str | Iterable[str]) -> set[str]:
        match classes:
            case str():
                return set(classes.split())

            case _:
                return set(classes)


class ContainerElement(Element):
    def __init__(self, body: str | Element | Sequence[str | Element], **attrs):
        self.body = body
        super().__init__(**attrs)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False

        return self.body == other.body

    def render(self):
        return Markup(
            f"<{self.tag} {self._build_attrs()}>{Markup.escape(self.body)}</{self.tag}>"
        )

    def _create_clone(self, **attrs) -> "ContainerElement":
        return type(self)(self.body, __clone__=True, **attrs)


class AElement(ContainerElement):
    tag = "a"


class InputElement(Element):
    tag = "input"


class ButtonElement(ContainerElement):
    tag = "button"


class LabelElement(ContainerElement):
    tag = "label"



import operator

from wordlette.utils.integer_infinity import Infinity


class Range:
    """An inclusive range implementation that can be unbounded on either end."""

    def __init__(self, start: int | None = None, end: int | None = None, step: int = 1):
        if step == 0:
            raise ValueError("step cannot be 0")

        self._end = Infinity(step < 0) if end is None else end
        self._start = Infinity(step >= 0) if start is None else start
        self._step = step

        self._contains_check = self._create_contains_check()

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def step(self):
        return self._step

    def __contains__(self, n: int):
        return self._contains_check(n)

    def _create_contains_check(self):
        op = operator.ge if self.step > 0 else operator.le

        def contains(n: int):
            if not op(n, self.start):
                return False

            if not op(self.end, n):
                return False

            offset_n = n if isinstance(self.start, Infinity) else n - self.start
            return offset_n % self.step == 0

        return contains



from wordlette.cms.states.setup.enums import SetupCategory
from wordlette.cms.states.setup.route_types import SetupRoute
from wordlette.cms.themes import Template
from wordlette.requests import Request


class SetupComplete(SetupRoute, setup_category=SetupCategory.NoCategory):
    path = "/setup-complete"

    async def get_setup_page(self, _: Request.Get):
        return Template(
            "setup-complete.html",
            title="Wordlette",
            subtitle="Setup Complete",
        )



from typing import _AnnotatedAlias


class AnnotatedAggregator(_AnnotatedAlias, _root=True):
    def __ror__(self, other):
        self.__origin__ = other | self.__origin__
        return self

    def __class_getitem__(cls, item):
        origin, *args = item
        if origin is None:
            origin = type(None)

        return cls(origin, tuple(args))



from wordlette.base_exceptions import BaseWordletteException


class ConfigCannotCreateFiles(BaseWordletteException):
    """Raised when a config handler cannot be used to create a file."""

    name = "Cannot Create Config Files"


class ConfigCannotLoadFiles(BaseWordletteException):
    """Raised when a config handler cannot be used to load a file."""

    name = "Cannot Load Config Files"



def run():
    import sys

    match (*map(str.casefold, sys.argv),):
        case (_, "serve", *args):
            _start_server(debug="--debug" in args or "-d" in args)

        case _:
            raise RuntimeError("Invalid command")


def _start_server(**settings):
    from wordlette.cms.app_bootstrap import create_app
    import uvicorn

    uvicorn.run(create_app(**settings), port=8000, log_level="info")



from wordlette.base_exceptions import BaseWordletteException


class MissingRoutePath(BaseWordletteException):
    """Raised when a route is missing a path."""

    name = "Route Path is Missing"


class NoRouteHandlersFound(BaseWordletteException):
    """Raised when a route has no registered method handlers."""

    name = "No Route Handlers Found"


class InvalidExtensionOrConstructor(BaseWordletteException):
    """Raised when an invalid extension or constructor is given."""

    name = "Invalid Extension or Constructor"


class ConfigFileNotFound(BaseWordletteException):
    """Raised when a config file cannot be found."""

    name = "Config File Not Found"


class CannotHandleInconsistentTypes(BaseWordletteException):
    """Raised when a route handler is registered to handle inconsistent types."""

    name = "Route Handlers Cannot Handle Inconsistent Types"



from collections.abc import Set
from typing import Type, Iterator

from wordlette.requests import Request


class MethodsCollection(Set[Type[Request]]):
    def __init__(self, *methods: Type[Request]):
        self._methods = set(methods)

    @property
    def names(self):
        return [method.name for method in self]

    def __contains__(self, request: Request | Type[Request]) -> bool:
        request_type = type(request) if isinstance(request, Request) else request
        return request_type in self._methods

    def __iter__(self) -> Iterator[Type[Request]]:
        return iter(self._methods)

    def __len__(self) -> int:
        return len(self._methods)

    def __repr__(self):
        return f"<{type(self).__name__} {{{', '.join(sorted(method.name for method in self))}}}>"



from wordlette.cms.states.setup.route_types import SetupRoute
from wordlette.cms.themes import ThemeManager
from wordlette.middlewares.router_middleware import RouteManager
from wordlette.state_machines import State
from wordlette.utils.dependency_injection import AutoInject, inject


class Setup(State, AutoInject):
    async def enter_state(
        self,
        route_manager: RouteManager @ inject,
        theme_manager: ThemeManager @ inject,
    ):
        self._load_route_modules()
        theme_manager.set_theme(theme_manager.wordlette_res / "themes" / "setup")
        SetupRoute.register_routes(route_manager.router)

    def _load_route_modules(self):
        # noinspection PyUnresolvedReferences
        import wordlette.cms.states.setup.routes



from typing import Any, Mapping

from bevy import inject, dependency
from starlette.background import BackgroundTask
from starlette.responses import HTMLResponse, Response

from wordlette.cms.themes.theme_managers import ThemeManager


class Template(Response):
    def __init__(
        self,
        _template_name: str,
        _status_code: int = 200,
        *,
        _headers: Mapping[str, str] | None = None,
        _media_type: str | None = None,
        _background: BackgroundTask | None = None,
        **context: Any,
    ):
        self.name = _template_name
        self.context = context
        self.status_code = _status_code
        self._headers = _headers
        self.media_type = _media_type
        self.background = _background

    def __call__(self, *args, **kwargs):
        return HTMLResponse(
            self._render(),
            self.status_code,
            self.headers,
            self.media_type,
            self.background,
        )(*args, **kwargs)

    @inject
    def _render(self, theme_manager: ThemeManager = dependency()) -> str:
        return theme_manager.render_template(self.name, **self.context)



import logging
import traceback
from dataclasses import dataclass
from io import StringIO

from bevy import inject, dependency
from starlette.exceptions import HTTPException
from starlette.types import Scope

from wordlette.cms.themes import Template, ThemeManager
from wordlette.core.events import LifespanStartupEvent
from wordlette.events import Observer
from wordlette.extensions import Extension
from wordlette.middlewares.router_middleware import RouteManager
from wordlette.utils.options import Value

logger = logging.getLogger(__name__)


@dataclass
class ExceptionObject:
    name: str
    message: str
    stacktrace: str
    notes: list[str]


class ErrorPages(Extension, Observer):
    @inject
    async def on_startup(
        self, _: LifespanStartupEvent, route_manager: RouteManager = dependency()
    ):
        logger.debug("Adding error pages")
        route_manager.add_error_page(0, self._error_page)

    @inject
    def _error_page(
        self, status_code: int, scope: Scope, theme: ThemeManager = dependency()
    ) -> Template:
        match theme.find_template(f"errors/{status_code}.html"):
            case Value():
                template_name = f"errors/{status_code}.html"

            case _:
                template_name = "errors/default.html"

        exception_name = None
        if "exception" in scope:
            exception = scope["exception"]
            exception_name = type(exception).__name__
            if isinstance(exception, HTTPException) and exception.detail is not None:
                exception_name = exception.detail
            elif hasattr(exception, "name") and exception.name is not None:
                exception_name = exception.name

        name = "An error was encountered"
        if exception_name:
            name = exception_name
        elif status_code == 404:
            name = "Page not found"

        return Template(
            template_name,
            name=name,
            title=f"{status_code} - {name}",
            path=scope["path"],
            status_code=status_code,
            exception=(
                ExceptionObject(
                    exception_name,
                    getattr(scope["exception"], "detail", str(scope["exception"])),
                    self._get_stacktrace(scope["exception"]),
                    getattr(scope["exception"], "__notes__", []),
                )
                if "exception" in scope
                else None
            ),
        )

    def _get_stacktrace(self, exception: Exception) -> str | None:
        io = StringIO()
        traceback.print_exception(exception, file=io)
        return io.getvalue()



from pathlib import Path
from typing import Annotated

from bevy import get_repository

from wordlette.cms.extensions.error_pages import ErrorPages
from wordlette.cms.states.setup import Setup
from wordlette.configs.managers import ConfigManager
from wordlette.core import WordletteApp
from wordlette.middlewares.router_middleware import RouterMiddleware
from wordlette.state_machines import StateMachine


def _get_config_handlers():
    from wordlette.configs import JsonHandler, TomlHandler, YamlHandler

    yield from (
        handler
        for handler in (
            YamlHandler,
            JsonHandler,
            TomlHandler,
        )
        if not hasattr(handler, "supported") or handler.supported()
    )


def _setup_repository():
    repository = get_repository()
    repository.set(
        Annotated[Path, "package-resources"], Path(__file__).parent / "resources"
    )

    repository.set(ConfigManager, ConfigManager(_get_config_handlers()))


def create_app(**settings):
    _setup_repository()
    return WordletteApp(
        extensions=[ErrorPages],
        middleware=[RouterMiddleware],
        state_machine=StateMachine(Setup),
        settings=settings,
    )



from typing import Iterable, Any

from wordlette.forms import Field
from wordlette.forms.elements import Element, LabelElement
from wordlette.forms.field_types import Button, not_set


class FormView:
    def __init__(
        self,
        fields: dict[str, Field],
        buttons: Iterable[Button],
        values: dict[str, Any],
        errors: dict[str, Exception],
    ):
        self._buttons = None
        self._fields = None
        self._labels = None

        self.errors = errors
        self.values = {
            fields[name].attrs["name"]: value for name, value in values.items()
        }

        self.raw_buttons = buttons
        self.raw_fields = fields

    @property
    def buttons(self) -> list[Element]:
        if self._buttons is None:
            self._buttons = self._compose_buttons(self.raw_buttons)

        return self._buttons

    @property
    def fields(self) -> dict[str, Element]:
        if self._fields is None:
            self._fields = self._compose_fields(self.raw_fields, self.values)

        return self._fields

    @property
    def labels(self) -> dict[str, LabelElement]:
        if self._labels is None:
            self._labels = self._compose_labels(self.raw_fields)

        return self._labels

    def _compose_buttons(self, buttons: Iterable[Button]) -> list[Element]:
        return [button.compose() for button in buttons]

    def _compose_fields(
        self, fields: dict[str, Field], values: dict[str, Any]
    ) -> dict[str, Element]:
        return {
            field.attrs["name"]: field.compose(values.get(field.attrs["name"], not_set))
            for field in fields.values()
        }

    def _compose_labels(self, fields: dict[str, Field]) -> dict[str, LabelElement]:
        return {field.attrs["name"]: field.compose_label() for field in fields.values()}



from typing import get_origin, get_args, Annotated, TypeVar

from bevy.options import Null, Value, Option
from bevy.provider_state import ProviderState
from bevy.providers import AnnotatedProvider

from wordlette.utils.annotated_aggregator import AnnotatedAggregator

T = TypeVar("T")


class AtAnnotationMCS(type):
    def __rmatmul__(self, other: T) -> T:
        return AnnotatedAggregator[other, self()]


class AtAnnotation(metaclass=AtAnnotationMCS):
    __match_args__ = ("strategy",)

    def __rmatmul__(self, other):
        return AnnotatedAggregator[other, self]


class AtProvider(AnnotatedProvider):
    def factory(self, key: T, cache: ProviderState) -> Option[T]:
        """Returns a callable that will get or construct an instance of the annotated type. If no instances exist in the
        repository matching the annotation, this will call the `Repository.get` method looking for the un-annotated
        type. That will attempt to instantiate an instance of the type if no providers have an instance cached.
        """
        match get_args(key):
            case (type_, AtAnnotation(strategy)):
                return Value(strategy(type_, cache))

            case Null(message):
                return Null(message)

    def supports(self, key, _) -> bool:
        if not get_origin(key) is Annotated:
            return False

        match get_args(key):
            case (_, AtAnnotation()):
                return True

            case _:
                return False



class Infinity(int):
    def __new__(cls, negative: bool = False):
        positive_inf = super().__new__(cls)
        positive_inf.negative = False

        negative_inf = super().__new__(cls)
        negative_inf.negative = True

        cls.__new__ = lambda _, neg=False: negative_inf if neg else positive_inf
        return cls.__new__(cls, negative)

    def __eq__(self, other):
        return isinstance(other, Infinity) and self.negative == other.negative

    def __lt__(self, other):
        return self.negative

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        return not self.negative

    def __ge__(self, other):
        return self == other or self > other

    def __neg__(self):
        return Infinity(not self.negative)

    def __repr__(self):
        return f"int({'-' * self.negative}Infinity)"
